@@ -15,7 +15,6 @@ import hashlib
 
 from oslo_serialization import jsonutils
 from oslo_utils import timeutils
-import six
 
 from keystonemiddleware.auth_token import _exceptions as exc
 from keystonemiddleware.auth_token import _memcache_crypt as memcache_crypt
@@ -28,7 +27,7 @@ def _hash_key(key):
     Using a known-length cache key is important to ensure that memcache
     maximum key length is not exceeded causing failures to validate.
     """
-    if isinstance(key, six.text_type):
+    if isinstance(key, str):
         # NOTE(morganfainberg): Ensure we are always working with a bytes
         # type required for the hasher. In python 2.7 it is possible to
         # get a text_type (unicode). In python 3.4 all strings are
@@ -53,8 +52,12 @@ class _EnvCachePool(object):
 class _CachePool(list):
     """A lazy pool of cache references."""
 
-    def __init__(self, memcached_servers, log):
+    def __init__(self, memcached_servers, log, arguments, tls_context=None):
         self._memcached_servers = memcached_servers
+        self._sasl_enabled = arguments.get("sasl_enabled", False)
+        self._username = arguments.get("username", None)
+        self._password = arguments.get("password", None)
+        self._tls_context = tls_context
         if not self._memcached_servers:
             log.warning(
                 "Using the in-process token cache is deprecated as of the "
@@ -74,8 +77,14 @@ class _CachePool(list):
         except IndexError:
             # the pool is empty, so we need to create a new client
             if self._memcached_servers:
-                import memcache
-                c = memcache.Client(self._memcached_servers, debug=0)
+                if self._sasl_enabled:
+                    import bmemcached
+                    c = bmemcached.Client(self._memcached_servers,
+                                          self._username, self._password)
+                else:
+                    import pymemcache
+                    c = pymemcache.HashClient(self._memcached_servers,
+                                              tls_context=self._tls_context)
             else:
                 c = _FakeClient()
 
@@ -89,16 +98,30 @@ class _MemcacheClientPool(object):
     """An advanced memcached client pool that is eventlet safe."""
 
     def __init__(self, memcache_servers, arguments, **kwargs):
-        # NOTE(sileht): This will import python-memcached, we don't want
-        # it as hard dependency, so lazy load it.
-        from oslo_cache import _memcache_pool
-        self._pool = _memcache_pool.MemcacheClientPool(memcache_servers,
-                                                       arguments,
-                                                       **kwargs)
+        # NOTE(sileht): This will import python-memcached and
+        # python-binary-memcached , we don't want it as hard
+        # dependency, so lazy load it.
+        self._sasl_enabled = arguments.pop("sasl_enabled", False)
+        if self._sasl_enabled:
+            from oslo_cache import _bmemcache_pool
+            self._pool = _bmemcache_pool.BMemcacheClientPool(memcache_servers,
+                                                             arguments,
+                                                             **kwargs)
+        else:
+            from oslo_cache import _memcache_pool
+            arguments.pop("username", None)
+            arguments.pop("password", None)
+            self._pool = _memcache_pool.MemcacheClientPool(memcache_servers,
+                                                           arguments,
+                                                           **kwargs)
 
     @contextlib.contextmanager
     def reserve(self):
-        yield self._pool.get()
+        # NOTE(morgan): We must use "acquire" if we want all the added context
+        # manager logic that places the connection back into the pool at the
+        # end of it's use.
+        with self._pool.acquire() as client:
+            yield client
 
 
 class TokenCache(object):
@@ -119,17 +142,21 @@ class TokenCache(object):
     _CACHE_KEY_TEMPLATE = 'tokens/%s'
 
     def __init__(self, log, cache_time=None,
-                 env_cache_name=None, memcached_servers=None,
-                 use_advanced_pool=False, dead_retry=None, socket_timeout=None,
+                 env_cache_name=None, memcached_servers=None, tls_context=None,
+                 use_advanced_pool=True, dead_retry=None, socket_timeout=None,
                  **kwargs):
         self._LOG = log
         self._cache_time = cache_time
         self._env_cache_name = env_cache_name
         self._memcached_servers = memcached_servers
+        self._tls_context = tls_context
         self._use_advanced_pool = use_advanced_pool
         self._arguments = {
             'dead_retry': dead_retry,
-            'socket_timeout': socket_timeout
+            'socket_timeout': socket_timeout,
+            'sasl_enabled': kwargs.pop("sasl_enabled", False),
+            'username': kwargs.pop("username", None),
+            'password': kwargs.pop("password", None)
         }
         self._memcache_pool_options = kwargs
 
@@ -146,7 +173,15 @@ class TokenCache(object):
                                        **self._memcache_pool_options)
 
         else:
-            return _CachePool(self._memcached_servers, self._LOG)
+            if not self._use_advanced_pool:
+                self._LOG.warning(
+                    "Using the eventlet-unsafe cache pool is deprecated."
+                    "It is recommended to use eventlet-safe cache pool"
+                    "implementation from oslo.cache. This can be enabled"
+                    "through config option memcache_use_advanced_pool = True")
+
+            return _CachePool(self._memcached_servers, self._LOG,
+                              self._arguments, tls_context=self._tls_context)
 
     def initialize(self, env):
         if self._initialized:
@@ -225,11 +260,15 @@ class TokenCache(object):
         if serialized is None:
             return None
 
-        if isinstance(serialized, six.text_type):
+        if isinstance(serialized, str):
             serialized = serialized.encode('utf8')
         data = self._deserialize(serialized, context)
 
-        if not isinstance(data, six.string_types):
+        if data is None:
+            # In case decryption fails, e.g. data corrupted in memcached.
+            return None
+
+        if not isinstance(data, str):
             data = data.decode('utf-8')
 
         return jsonutils.loads(data)
@@ -237,7 +276,7 @@ class TokenCache(object):
     def set(self, token_id, data):
         """Store value into memcache."""
         data = jsonutils.dumps(data)
-        if isinstance(data, six.text_type):
+        if isinstance(data, str):
             data = data.encode('utf-8')
 
         cache_key, context = self._get_cache_key(token_id)
@@ -262,9 +301,9 @@ class SecureTokenCache(TokenCache):
                     'memcache_security_strategy is defined')
             raise exc.ConfigurationError(msg)
 
-        if isinstance(security_strategy, six.string_types):
+        if isinstance(security_strategy, str):
             security_strategy = security_strategy.encode('utf-8')
-        if isinstance(secret_key, six.string_types):
+        if isinstance(secret_key, str):
             secret_key = secret_key.encode('utf-8')
 
         self._security_strategy = security_strategy
